import colorsys
import logging
import math
import sys
from collections import deque
from datetime import timedelta
from enum import Enum

import numpy as np
import numpy.typing as npt
import pygame
import pygame_gui
import yaml
from scipy.spatial import KDTree

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.counters import Counter, CookingCounter
from overcooked_simulator.game_items import (
    Item,
    CookingEquipment,
    Plate,
)
from overcooked_simulator.gui_2d_vis.game_colors import BLUE
from overcooked_simulator.gui_2d_vis.game_colors import colors, Color
from overcooked_simulator.order import Order
from overcooked_simulator.overcooked_environment import (
    Action,
    ActionType,
    InterActionData,
)
from overcooked_simulator.simulation_runner import Simulator

USE_PLAYER_COOK_SPRITES = True
SHOW_INTERACTION_RANGE = False
SHOW_COUNTER_CENTERS = False


class MenuStates(Enum):
    Start = "Start"
    Game = "Game"
    End = "End"


def create_polygon(n, length):
    if n == 1:
        return np.array([0, 0])

    vector = np.array([length, 0])
    angle = (2 * np.pi) / n

    rot_matrix = np.array(
        [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
    )

    vecs = [vector]
    for i in range(n - 1):
        vector = np.dot(rot_matrix, vector)
        vecs.append(vector)

    return vecs


log = logging.getLogger(__name__)


class PlayerKeySet:
    """Set of keyboard keys for controlling a player.
    First four keys are for movement. Order: Down, Up, Left, Right.
    5th key is for interacting with counters.
    6th key ist for picking up things or dropping them.

    """

    def __init__(self, player_name: str, keys: list[pygame.key]):
        """Creates a player key set which contains information about which keyboard keys control the player.
        Movement keys in the following order: Down, Up, Left, Right

        Args:
            player_name: The name of the player to control.
            keys: The keys which control this player in the following order: Down, Up, Left, Right, Interact, Pickup.
        """
        self.name = player_name
        self.player_keys = keys
        self.move_vectors = [[-1, 0], [1, 0], [0, -1], [0, 1]]
        self.key_to_movement = {
            key: vec for (key, vec) in zip(self.player_keys[:-2], self.move_vectors)
        }
        self.interact_key = self.player_keys[-2]
        self.pickup_key = self.player_keys[-1]


class PyGameGUI:
    """Visualisation of the overcooked environment and reading keyboard inputs using pygame."""

    def __init__(
        self,
        simulator: Simulator,
        player_names: list[str],
        player_keys: list[pygame.key],
    ):
        self.game_screen = None
        self.FPS = 60
        self.simulator: Simulator = simulator
        self.running = True

        self.player_names = player_names
        self.player_keys = player_keys

        self.player_key_sets: list[PlayerKeySet] = [
            PlayerKeySet(player_name, keys)
            for player_name, keys in zip(
                self.player_names, self.player_keys[: len(self.player_names)]
            )
        ]

        # TODO cache loaded images?
        with open(ROOT_DIR / "gui_2d_vis" / "visualization.yaml", "r") as file:
            self.visualization_config = yaml.safe_load(file)

        self.screen_margin = self.visualization_config["GameWindow"]["screen_margin"]
        self.min_width = self.visualization_config["GameWindow"]["min_width"]
        self.min_height = self.visualization_config["GameWindow"]["min_height"]

        self.buttons_width = self.visualization_config["GameWindow"]["buttons_width"]
        self.buttons_height = self.visualization_config["GameWindow"]["buttons_height"]

        self.order_bar_height = self.visualization_config["GameWindow"][
            "order_bar_height"
        ]

        self.window_width = self.min_width
        self.window_height = self.min_height

        self.main_window = pygame.display.set_mode(
            (self.window_width, self.window_height)
        )

        # self.game_width, self.game_height = 0, 0

        self.images_path = ROOT_DIR / "pygame_gui" / "images"

        self.image_cache_dict = {}

        self.menu_state = MenuStates.Start
        self.manager: pygame_gui.UIManager

    def get_window_sizes(self, state: dir):
        counter_positions = np.array([c.pos for c in state["counters"]])
        kitchen_width = counter_positions[:, 0].max() + 0.5
        kitchen_height = counter_positions[:, 1].max() + 0.5
        if self.visualization_config["GameWindow"]["WhatIsFixed"] == "window_width":
            game_width = self.visualization_config["GameWindow"]["size"]
            kitchen_aspect_ratio = kitchen_height / kitchen_width
            game_height = int(game_width * kitchen_aspect_ratio)
            grid_size = int(game_width / self.simulator.env.kitchen_width)

        elif self.visualization_config["GameWindow"]["WhatIsFixed"] == "window_height":
            game_height = self.visualization_config["GameWindow"]["size"]
            kitchen_aspect_ratio = kitchen_width / kitchen_height
            game_width = int(game_height * kitchen_aspect_ratio)
            grid_size = int(game_width / self.simulator.env.kitchen_width)

        elif self.visualization_config["GameWindow"]["WhatIsFixed"] == "grid":
            grid_size = self.visualization_config["GameWindow"]["size"]
            game_width, game_height = (
                kitchen_width * grid_size,
                kitchen_height * grid_size,
            )

        else:
            game_width, game_height = 0, 0
            grid_size = 0

        window_width, window_height = (
            game_width + (2 * self.screen_margin),
            game_height + (2 * self.screen_margin),  # bar with orders
        )

        window_width = max(window_width, self.min_width)
        window_height = max(window_height, self.min_height)
        return (
            int(window_width),
            int(window_height),
            game_width,
            game_height,
            grid_size,
        )

    def create_player_colors(self) -> list[Color]:
        number_player = len(self.simulator.env.players)
        hue_values = np.linspace(0, 1, number_player + 1)

        colors_vec = np.array([col for col in colors.values()])

        tree = KDTree(colors_vec)

        color_names = list(colors.keys())

        player_colors = []
        for hue in hue_values:
            rgb = colorsys.hsv_to_rgb(hue, 1, 1)
            query_color = np.array([int(c * 255) for c in rgb])
            _, index = tree.query(query_color, k=1)
            player_colors.append(color_names[index])

        return player_colors

    def send_action(self, action: Action):
        """Sends an action to the game environment.

        Args:
            action: The action to be sent. Contains the player, action type and move direction if action is a movement.
        """
        self.simulator.enter_action(action)

    def handle_keys(self):
        """Handles keyboard inputs. Sends action for the respective players. When a key is held down, every frame
        an action is sent in this function.
        """
        keys = pygame.key.get_pressed()
        for player_idx, key_set in enumerate(self.player_key_sets):
            relevant_keys = [keys[k] for k in key_set.player_keys]
            if any(relevant_keys[:-2]):
                move_vec = np.zeros(2)
                for idx, pressed in enumerate(relevant_keys[:-2]):
                    if pressed:
                        move_vec += key_set.move_vectors[idx]
                if np.linalg.norm(move_vec) != 0:
                    move_vec = move_vec / np.linalg.norm(move_vec)

                action = Action(key_set.name, ActionType.MOVEMENT, move_vec)
                self.send_action(action)

    def handle_key_event(self, event):
        """Handles key events for the pickup and interaction keys. Pickup is a single action,
        for interaction keydown and keyup is necessary, because the player has to be able to hold
        the key down.

        Args:
            event: Pygame event for extracting the key action.
        """
        for key_set in self.player_key_sets:
            if event.key == key_set.pickup_key and event.type == pygame.KEYDOWN:
                action = Action(key_set.name, ActionType.PUT, "pickup")
                self.send_action(action)

            if event.key == key_set.interact_key:
                if event.type == pygame.KEYDOWN:
                    action = Action(
                        key_set.name, ActionType.INTERACT, InterActionData.START
                    )
                    self.send_action(action)
                elif event.type == pygame.KEYUP:
                    action = Action(
                        key_set.name, ActionType.INTERACT, InterActionData.STOP
                    )
                    self.send_action(action)

    def draw_background(self):
        """Visualizes a game background."""
        block_size = self.grid_size // 2  # Set the size of the grid block
        self.game_screen.fill(
            colors[self.visualization_config["Kitchen"]["ground_tiles_color"]]
        )
        for x in range(0, self.window_width, block_size):
            for y in range(0, self.window_height, block_size):
                rect = pygame.Rect(x, y, block_size, block_size)
                pygame.draw.rect(
                    self.game_screen,
                    self.visualization_config["Kitchen"]["background_lines"],
                    rect,
                    1,
                )

    def draw_image(
        self, img_path, size, pos, rot_angle=0, screen: pygame.Surface = None
    ):
        cache_entry = f"{img_path}"
        if cache_entry in self.image_cache_dict.keys():
            image = self.image_cache_dict[cache_entry]
        else:
            image = pygame.image.load(
                ROOT_DIR / "gui_2d_vis" / img_path
            ).convert_alpha()
            self.image_cache_dict[cache_entry] = image

        image = pygame.transform.scale(image, (size, size))
        if rot_angle != 0:
            image = pygame.transform.rotate(image, rot_angle)
        rect = image.get_rect()
        rect.center = pos

        if screen is None:
            self.game_screen.blit(image, rect)
        else:
            screen.blit(image, rect)

    def draw_players(self, state):
        """Visualizes the players as circles with a triangle for the facing direction.
        If the player holds something in their hands, it is displayed

        Args:
            state: The game state returned by the environment.
        """
        for p_idx, player in enumerate(state["players"].values()):
            pos = player.pos * self.grid_size

            if USE_PLAYER_COOK_SPRITES:
                img_path = self.visualization_config["Cook"]["parts"][0]["path"]
                rel_x, rel_y = player.facing_direction
                angle = -np.rad2deg(math.atan2(rel_y, rel_x)) + 90
                size = (
                    self.visualization_config["Cook"]["parts"][0]["size"]
                    * self.grid_size
                )
                self.draw_image(img_path, size, pos, angle)

            else:
                size = player.radius * self.grid_size
                color1 = self.player_colors[p_idx]
                color2 = colors["white"]

                pygame.draw.circle(self.game_screen, color2, pos, size)
                pygame.draw.circle(self.game_screen, BLUE, pos, size, width=1)
                pygame.draw.circle(self.game_screen, colors[color1], pos, size // 2)

                facing = player.facing_direction
                pygame.draw.polygon(
                    self.game_screen,
                    BLUE,
                    (
                        (
                            pos[0] + (facing[1] * 0.1 * self.grid_size),
                            pos[1] - (facing[0] * 0.1 * self.grid_size),
                        ),
                        (
                            pos[0] - (facing[1] * 0.1 * self.grid_size),
                            pos[1] + (facing[0] * 0.1 * self.grid_size),
                        ),
                        pos + (facing * 0.5 * self.grid_size),
                    ),
                )

            if SHOW_INTERACTION_RANGE:
                pygame.draw.circle(
                    self.game_screen,
                    BLUE,
                    player.facing_point * self.grid_size,
                    player.interaction_range * self.grid_size,
                    width=1,
                )
                pygame.draw.circle(
                    self.game_screen,
                    colors["red1"],
                    player.facing_point * self.grid_size,
                    4,
                )
                pygame.draw.circle(
                    self.game_screen, colors["red1"], player.facing_point, 4
                )

            if player.holding is not None:
                holding_item_pos = (player.pos * self.grid_size) + (
                    20 * player.facing_direction
                )
                self.draw_item(holding_item_pos, player.holding)

            if player.current_nearest_counter:
                counter: Counter = player.current_nearest_counter
                pos = counter.pos * self.grid_size
                pygame.draw.rect(
                    self.game_screen,
                    colors[self.player_colors[p_idx]],
                    rect=pygame.Rect(
                        pos[0] - (self.grid_size // 2),
                        pos[1] - (self.grid_size // 2),
                        self.grid_size,
                        self.grid_size,
                    ),
                    width=2,
                )

    def draw_thing(
        self,
        pos: npt.NDArray[float],
        parts: list[dict[str]],
        scale: float = 1.0,
        screen: pygame.Surface = None,
    ):
        """Draws an item, based on its visual parts specified in the visualization config.

        Args:
            pos: Where to draw the item parts.
            parts: The visual parts to draw.
            scale: Rescale the item by this factor.
        """

        if screen is None:
            screen = self.game_screen

        for part in parts:
            part_type = part["type"]
            match part_type:
                case "image":
                    if "center_offset" in part:
                        d = np.array(part["center_offset"]) * self.grid_size
                        pos += d

                    self.draw_image(
                        part["path"],
                        part["size"] * scale * self.grid_size,
                        pos,
                        screen=screen,
                    )
                case "rect":
                    height = part["height"] * self.grid_size
                    width = part["width"] * self.grid_size
                    color = part["color"]
                    if "center_offset" in part:
                        dx, dy = np.array(part["center_offset"]) * self.grid_size
                        rect = pygame.Rect(pos[0] + dx, pos[1] + dy, height, width)
                        pygame.draw.rect(screen, color, rect)
                    else:
                        rect = pygame.Rect(
                            pos[0] - (height / 2),
                            pos[1] - (width / 2),
                            height,
                            width,
                        )
                    pygame.draw.rect(screen, color, rect)
                case "circle":
                    radius = part["radius"] * self.grid_size
                    color = colors[part["color"]]
                    if "center_offset" in part:
                        pygame.draw.circle(
                            self.game_screen,
                            color,
                            pos + (np.array(part["center_offset"]) * self.grid_size),
                            radius,
                        )
                    else:
                        pygame.draw.circle(screen, color, pos, radius)

    def draw_item(
        self,
        pos: npt.NDArray[float],
        item: Item,
        scale: float = 1.0,
        plate=False,
        screen=None,
    ):
        """Visualization of an item at the specified position. On a counter or in the hands of the player.
        The visual composition of the item is read in from visualization.yaml file, where it is specified as
        different parts to be drawn.

        Args:
            pos: The position of the item to draw.
            item: The item do be drawn in the game.
            scale: Rescale the item by this factor.
            screen: the pygame screen to draw on.
            plate: item is on a plate (soup are is different on a plate and pot)
        """

        if not isinstance(item, list):
            if item.name in self.visualization_config:
                item_key = item.name
                if "Soup" in item.name and plate:
                    item_key += "Plate"
                self.draw_thing(
                    pos,
                    self.visualization_config[item_key]["parts"],
                    scale=scale,
                    screen=screen,
                )

        if isinstance(item, (Item, Plate)) and item.progress_percentage > 0.0:
            self.draw_progress_bar(pos, item.progress_percentage)

        if isinstance(item, CookingEquipment) and item.content_list:
            if (
                item.content_ready
                and item.content_ready.name in self.visualization_config
            ):
                self.draw_thing(
                    pos,
                    self.visualization_config[item.content_ready.name]["parts"],
                    screen=screen,
                )
            else:
                triangle_offsets = create_polygon(len(item.content_list), length=10)
                scale = 1 if len(item.content_list) == 1 else 0.6
                for idx, o in enumerate(item.content_list):
                    self.draw_item(
                        pos + triangle_offsets[idx],
                        o,
                        scale=scale,
                        plate=isinstance(item, Plate),
                        screen=screen,
                    )

        # if isinstance(item, Meal):
        #     if item.finished:
        #         if item.name in self.visualization_config:
        #             self.draw_thing(pos, self.visualization_config[item.name]["parts"])
        #     else:
        #         for idx, o in enumerate(item.parts):
        #             triangle_offsets = create_polygon(len(item.parts), length=10)
        #             self.draw_item(pos + triangle_offsets[idx], o, scale=0.6)

    def draw_progress_bar(self, pos, percent, screen=None):
        """Visualize progress of progressing item as a green bar under the item."""
        bar_height = self.grid_size * 0.2
        progress_width = percent * self.grid_size
        progress_bar = pygame.Rect(
            pos[0] - (self.grid_size / 2),
            pos[1] - (self.grid_size / 2) + self.grid_size - bar_height,
            progress_width,
            bar_height,
        )
        if screen is None:
            pygame.draw.rect(self.game_screen, colors["green1"], progress_bar)
        else:
            pygame.draw.rect(screen, colors["green1"], progress_bar)

    def draw_counter(self, counter):
        """Visualization of a counter at its position. If it is occupied by an item, it is also shown.
        The visual composition of the counter is read in from visualization.yaml file, where it is specified as
        different parts to be drawn.

        Args:
            counter: The counter to visualize.
        """

        pos = counter.pos * self.grid_size
        self.draw_thing(pos, self.visualization_config["Counter"]["parts"])
        if str(counter) in self.visualization_config:
            self.draw_thing(pos, self.visualization_config[str(counter)]["parts"])
        else:
            self.draw_thing(
                pos,
                self.visualization_config[
                    counter.name
                    if isinstance(counter, CookingCounter)
                    else counter.__class__.__name__
                ]["parts"],
            )

        if counter.occupied_by is not None:
            # Multiple plates on plate return:
            if isinstance(counter.occupied_by, (list, deque)):
                with self.simulator.env.lock:
                    for i, o in enumerate(counter.occupied_by):
                        self.draw_item(np.abs([pos[0], pos[1] - (i * 3)]), o)
            # All other items:
            else:
                self.draw_item(pos, counter.occupied_by)

    def draw_counters(self, state):
        """Visualizes the counters in the environment.

        Args:
            state: The game state returned by the environment.
        """
        for counter in state["counters"]:
            self.draw_counter(counter)
            if SHOW_COUNTER_CENTERS:
                pygame.draw.circle(self.game_screen, colors["green1"], counter.pos, 3)

    def update_score_label(self, state):
        score = state["score"]
        self.score_label.set_text(f"Score {score}")

    def update_conclusion_label(self, state):
        score = state["score"]
        self.conclusion_label.set_text(f"Your final score is {score}. Hurray!")

    def update_remaining_time(self, remaining_time: timedelta):
        hours, rem = divmod(remaining_time.seconds, 3600)
        minutes, seconds = divmod(rem, 60)
        display_time = f"{minutes}:{'%02d' % seconds}"
        self.timer_label.set_text(f"Time remaining: {display_time}")

    def draw_orders(self, state):
        orders_width = self.game_width - 100
        orders_height = self.screen_margin

        order_screen = pygame.Surface(
            (orders_width, orders_height),
        )

        bg_color = colors[self.visualization_config["GameWindow"]["background_color"]]
        pygame.draw.rect(order_screen, bg_color, order_screen.get_rect())

        order_rects_start = (orders_height // 2) - (self.grid_size // 2)
        with self.simulator.env.lock:
            for idx, order in enumerate(state["orders"]):
                order: Order
                order_upper_left = [
                    order_rects_start + idx * self.grid_size * 1.2,
                    order_rects_start,
                ]
                pygame.draw.rect(
                    order_screen,
                    colors["red"],
                    pygame.Rect(
                        order_upper_left[0],
                        order_upper_left[1],
                        self.grid_size,
                        self.grid_size,
                    ),
                    width=2,
                )
                center = np.array(order_upper_left) + np.array(
                    [self.grid_size / 2, self.grid_size / 2]
                )
                self.draw_thing(
                    center,
                    self.visualization_config["Plate"]["parts"],
                    screen=order_screen,
                )
                self.draw_item(
                    center,
                    order.meal,
                    plate=True,
                    screen=order_screen,
                )
                order_done_seconds = (
                    (order.start_time + order.max_duration) - state["env_time"]
                ).total_seconds()

                percentage = order_done_seconds / order.max_duration.total_seconds()
                self.draw_progress_bar(center, percentage, screen=order_screen)

        orders_rect = order_screen.get_rect()
        orders_rect.center = [
            self.screen_margin + (orders_width // 2),
            orders_height // 2,
        ]
        self.main_window.blit(order_screen, orders_rect)

    def draw(self, state):
        """Main visualization function.

        Args:
            state: The game state returned by the environment.
        """

        self.draw_background()

        self.draw_counters(state)
        self.draw_players(state)
        self.manager.draw_ui(self.main_window)
        self.update_remaining_time(state["remaining_time"])

        self.draw_orders(state)
        self.update_score_label(state)

    def init_ui_elements(self):
        self.manager = pygame_gui.UIManager((self.window_width, self.window_height))
        self.manager.get_theme().load_theme(ROOT_DIR / "gui_2d_vis" / "gui_theme.json")

        self.start_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (self.window_width // 2) - self.buttons_width // 2,
                    (self.window_height / 2) - self.buttons_height // 2,
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Start Game",
            manager=self.manager,
        )
        self.start_button.can_hover()

        self.quit_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (self.window_width - self.buttons_width),
                    0,
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Quit Game",
            manager=self.manager,
            object_id="#quit_button",
        )
        self.quit_button.can_hover()

        self.finished_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (self.window_width - self.buttons_width),
                    (self.window_height - self.buttons_height),
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Finish round",
            manager=self.manager,
        )
        self.finished_button.can_hover()

        self.back_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (0),
                    (self.window_height - self.buttons_height),
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Back to menu",
            manager=self.manager,
        )
        self.back_button.can_hover()

        self.score_label = pygame_gui.elements.UILabel(
            text=f"Score: _",
            relative_rect=pygame.Rect(
                (
                    (0),
                    self.window_height - self.screen_margin,
                ),
                (self.screen_margin * 2, self.screen_margin),
            ),
            manager=self.manager,
            object_id="#score_label",
        )

        layout_file_paths = [
            str(p.name)
            for p in (ROOT_DIR / "game_content" / "layouts").glob("*.layout")
        ]
        assert len(layout_file_paths) != 0, "No layout files."
        dropdown_width, dropdown_height = 200, 40
        self.layout_selection = pygame_gui.elements.UIDropDownMenu(
            relative_rect=pygame.Rect(
                (
                    0,
                    0,
                ),
                (dropdown_width, dropdown_height),
            ),
            manager=self.manager,
            options_list=layout_file_paths,
            starting_option=layout_file_paths[-1],
        )
        self.timer_label = pygame_gui.elements.UILabel(
            text="GAMETIME",
            relative_rect=pygame.Rect(
                (self.screen_margin, self.window_height - self.screen_margin),
                (self.game_width, self.screen_margin),
            ),
            manager=self.manager,
            object_id="#timer_label",
        )

        self.orders_label = pygame_gui.elements.UILabel(
            text="Orders:",
            relative_rect=pygame.Rect(0, 0, self.screen_margin, self.screen_margin),
            manager=self.manager,
            object_id="#orders_label",
        )

        self.conclusion_label = pygame_gui.elements.UILabel(
            text="Your final score was _",
            relative_rect=pygame.Rect(0, 0, self.window_width, self.window_height),
            manager=self.manager,
            object_id="#score_label",
        )

    def set_window_size(self, window_width, window_height, game_width, game_height):
        self.game_screen = pygame.Surface(
            (
                game_width,
                game_height,
            ),
        )
        self.main_window = pygame.display.set_mode(
            (
                window_width,
                window_height,
            )
        )

    def reset_window_size(self):
        self.window_width = self.min_width
        self.window_height = self.min_height
        self.game_width = 0
        self.game_height = 0
        self.set_window_size(self.min_width, self.min_height, 0, 0)
        self.init_ui_elements()

    def setup_simulation(self, config_path, layout_path):
        self.simulator = Simulator(config_path, layout_path, 600)
        number_player = len(self.player_names)
        for i in range(number_player):
            player_name = f"p{i}"
            self.simulator.register_player(player_name)
        self.simulator.start()
        (
            self.window_width,
            self.window_height,
            self.game_width,
            self.game_height,
            self.grid_size,
        ) = self.get_window_sizes(self.simulator.get_state())
        self.player_colors = self.create_player_colors()

    def manage_button_visibility(self):
        match self.menu_state:
            case MenuStates.Start:
                self.back_button.hide()
                self.quit_button.show()
                self.start_button.show()
                self.score_label.hide()
                self.finished_button.hide()
                self.layout_selection.show()
                self.timer_label.hide()
                self.orders_label.hide()
                self.conclusion_label.hide()
            case MenuStates.Game:
                self.start_button.hide()
                self.back_button.hide()
                self.score_label.show()
                self.score_label.show()
                self.finished_button.show()
                self.layout_selection.hide()
                self.timer_label.show()
                self.orders_label.show()
                self.conclusion_label.hide()
            case MenuStates.End:
                self.start_button.hide()
                self.back_button.show()
                self.score_label.hide()
                self.finished_button.hide()
                self.layout_selection.hide()
                self.timer_label.hide()
                self.orders_label.hide()
                self.conclusion_label.show()

    def start_button_press(self):
        self.menu_state = MenuStates.Game

        layout_path = (
            ROOT_DIR
            / "game_content"
            / "layouts"
            / self.layout_selection.selected_option
        )
        config_path = ROOT_DIR / "game_content" / "environment_config.yaml"

        self.setup_simulation(config_path, layout_path)

        self.set_window_size(*(self.get_window_sizes(self.simulator.get_state()))[:-1])

        self.init_ui_elements()
        log.debug("Pressed start button")

    def back_button_press(self):
        self.menu_state = MenuStates.Start
        self.reset_window_size()
        self.simulator.stop()
        log.debug("Pressed back button")

    def quit_button_press(self):
        self.simulator.stop()
        self.running = False
        log.debug("Pressed quit button")

    def finished_button_press(self):
        self.menu_state = MenuStates.End
        self.reset_window_size()
        self.simulator.stop()
        log.debug("Pressed finished button")

    def start_pygame(self):
        """Starts pygame and the gui loop. Each frame the game state is visualized and keyboard inputs are read."""
        log.debug(f"Starting pygame gui at {self.FPS} fps")
        pygame.init()
        pygame.font.init()

        pygame.display.set_caption("Simple Overcooked Simulator")

        clock = pygame.time.Clock()

        self.reset_window_size()
        self.init_ui_elements()
        self.manage_button_visibility()

        # Game loop
        self.running = True
        while self.running:
            try:
                time_delta = clock.tick(self.FPS) / 1000.0

                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        self.running = False

                    # UI Buttons:
                    if event.type == pygame_gui.UI_BUTTON_PRESSED:
                        match event.ui_element:
                            case self.start_button:
                                self.start_button_press()
                            case self.back_button:
                                self.back_button_press()
                            case self.finished_button:
                                self.finished_button_press()
                            case self.quit_button:
                                self.quit_button_press()

                        self.manage_button_visibility()

                    if (
                        event.type in [pygame.KEYDOWN, pygame.KEYUP]
                        and self.menu_state == MenuStates.Game
                    ):
                        self.handle_key_event(event)

                    self.manager.process_events(event)

                # drawing:

                state = self.simulator.get_state()

                self.main_window.fill(
                    colors[self.visualization_config["GameWindow"]["background_color"]]
                )
                self.manager.draw_ui(self.main_window)

                match self.menu_state:
                    case MenuStates.Start:
                        pass

                    case MenuStates.Game:
                        self.draw_background()

                        self.handle_keys()

                        if state["ended"]:
                            self.finished_button_press()
                            self.manage_button_visibility()
                        else:
                            self.draw(state)

                            game_screen_rect = self.game_screen.get_rect()
                            game_screen_rect.center = [
                                self.window_width // 2,
                                self.window_height // 2,
                            ]

                            self.main_window.blit(self.game_screen, game_screen_rect)

                    case MenuStates.End:
                        self.update_conclusion_label(state)

                self.manager.update(time_delta)
                pygame.display.flip()

            except KeyboardInterrupt:
                self.simulator.stop()
                pygame.quit()
                sys.exit()

        self.simulator.stop()
        pygame.quit()
        sys.exit()
