import time
from datetime import timedelta

import numpy as np
import pytest

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.counters import Counter, CuttingBoard
from overcooked_simulator.game_items import Item
from overcooked_simulator.overcooked_environment import (
    Action,
    Environment,
    ActionType,
    InterActionData,
)
from overcooked_simulator.simulation_runner import Simulator
from overcooked_simulator.utils import create_init_env_time

layouts_folder = ROOT_DIR / "game_content" / "layouts"


# TODO: TESTs are in absolute pixel coordinates still.


@pytest.fixture(autouse=True)
def test_file_availability():
    assert layouts_folder.is_dir(), "layouts folder does not exists"
    assert (
        layouts_folder / "basic.layout"
    ).is_file(), "basic layouts file does not exists"


def test_player_registration():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        200,
    )

    sim.register_player("player1", np.array([0, 0]))

    assert len(sim.env.players) != 0, "Wrong number of players"
    assert len(sim.env.players) == 1, "Wrong number of players"

    sim.register_player("player2", np.array([100, 100]))

    assert len(sim.env.players) == 2, "Wrong number of players"

    sim.register_player("player2", np.array([100, 100]))  # same player name
    assert len(sim.env.players) == 2, "Wrong number of players"

    sim.start()
    sim.stop()


def test_simulator_frequency():
    class TestEnv:
        def __init__(self):
            self.c = 0

        def step(self, passed_time):
            self.c += 1

        def reset_env_time(self):
            pass

    frequency = 2000
    running_time_seconds = 2

    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        frequency,
    )
    sim.env = TestEnv()  # Overwrite environment with a simple counting env

    sim.start()
    time.sleep(running_time_seconds)
    sim.stop()

    accepted_tolerance = 0.02
    lower = frequency * running_time_seconds * (1 - accepted_tolerance)
    upper = frequency * running_time_seconds * (1 + accepted_tolerance)
    assert lower < sim.env.c < upper, "Timing error in the environment at 1000hz"


def test_movement():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        200,
    )
    player_name = "p1"
    start_pos = np.array([1, 2])
    sim.register_player(player_name, start_pos)
    move_direction = np.array([1, 0])
    move_action = Action(player_name, ActionType.MOVEMENT, move_direction)
    do_moves_number = 6
    for i in range(do_moves_number):
        sim.enter_action(move_action)

    expected = start_pos + do_moves_number * (
        move_direction * sim.env.players[player_name].move_dist
    )

    assert np.isclose(
        np.linalg.norm(expected - sim.env.players[player_name].pos), 0
    ), "Should be here?"


def test_collision_detection():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        200,
    )
    counter_pos = np.array([1, 2])
    counter = Counter(counter_pos)
    sim.env.counters = [counter]
    sim.register_player("p1", np.array([1, 1]))
    sim.register_player("p2", np.array([1, 4]))  # same player name
    player1 = sim.env.players["p1"]
    player2 = sim.env.players["p2"]

    sim.start()
    try:
        assert not sim.env.detect_collision_counters(player1), "Should not collide"
        assert not sim.env.detect_player_collision(player1), "Should not collide yet."

        assert not sim.env.detect_collision(player1), "Does not collide yet."

        player1.move_abs(counter_pos)
        assert sim.env.detect_collision_counters(
            player1
        ), "Player and counter at same pos. Not detected."
        player2.move_abs(counter_pos)
        assert sim.env.detect_player_collision(
            player1
        ), "Players at same pos. Not detected."

        player1.move_abs(np.array([0, 0]))
        assert sim.env.detect_collision_world_bounds(
            player1
        ), "Player collides with world bounds."
    finally:
        sim.stop()


def test_player_reach():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        200,
    )

    counter_pos = np.array([2, 2])
    counter = Counter(counter_pos)
    sim.env.counters = [counter]
    sim.register_player("p1", np.array([2, 4]))
    player = sim.env.players["p1"]
    assert not player.can_reach(counter), "Player is too far away."

    do_moves_number = 30
    for i in range(do_moves_number):
        move_action = Action("p1", ActionType.MOVEMENT, np.array([0, -1]))
        sim.enter_action(move_action)
    assert player.can_reach(counter), "Player can reach counter?"


def test_pickup():
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        200,
    )

    counter_pos = np.array([2, 2])
    counter = Counter(counter_pos)
    counter.occupied_by = Item(name="Tomato", item_info=None)
    sim.env.counters = [counter]

    sim.register_player("p1", np.array([2, 3]))
    player = sim.env.players["p1"]

    move_down = Action("p1", ActionType.MOVEMENT, np.array([0, -1]))
    move_up = Action("p1", ActionType.MOVEMENT, np.array([0, 1]))
    pick = Action("p1", ActionType.PUT, "pickup")

    sim.enter_action(move_down)
    assert player.can_reach(counter), "Player can reach counter?"

    sim.enter_action(pick)
    assert player.holding is not None, "Player should have picked up tomato."
    assert player.holding.name == "Tomato", "Should be tomato."

    sim.enter_action(move_up)
    sim.enter_action(move_up)
    sim.enter_action(move_up)
    sim.enter_action(move_up)

    sim.enter_action(pick)
    assert (
        player.holding is not None
    ), "Player should be too far away to put tomato down."

    sim.enter_action(move_down)
    sim.enter_action(move_down)
    sim.enter_action(move_down)
    sim.enter_action(move_down)

    sim.enter_action(pick)

    assert player.holding is None, "Player should have put tomato down."
    assert (
        counter.occupied_by is not None and counter.occupied_by.name == "Tomato"
    ), "Tomato should be here now."


def test_processing():
    sim_frequency = 1000
    sim = Simulator(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        sim_frequency,
    )
    sim.start()
    try:
        counter_pos = np.array([2, 2])
        counter = CuttingBoard(
            counter_pos,
            transitions={"Tomato": {"seconds": 1, "result": "ChoppedTomato"}},
        )
        sim.env.counters.append(counter)

        tomato = Item(name="Tomato", item_info=None)
        sim.register_player("p1", np.array([2, 3]))
        player = sim.env.players["p1"]
        player.holding = tomato

        move = Action("p1", ActionType.MOVEMENT, np.array([0, -1]))
        pick = Action("p1", ActionType.PUT, "pickup")

        sim.enter_action(move)
        sim.enter_action(pick)

        hold_down = Action("p1", ActionType.INTERACT, InterActionData.START)
        sim.enter_action(hold_down)

        assert tomato.name != "ChoppedTomato", "Tomato is not finished yet."

        time.sleep(1)

        assert tomato.name == "ChoppedTomato", "Tomato should be finished."

        button_up = Action("p1", ActionType.INTERACT, InterActionData.STOP)
        sim.enter_action(button_up)
    finally:
        sim.stop()


def test_time_passed():
    np.random.seed(42)
    env = Environment(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        ROOT_DIR / "game_content" / "item_info.yaml",
    )
    env.reset_env_time()
    passed_time = timedelta(seconds=10)
    env.step(passed_time)
    assert (
        env.env_time == create_init_env_time() + passed_time
    ), "Env time needs to be updated via the step function"

    passed_time_2 = timedelta(seconds=12)
    env.step(passed_time_2)
    assert (
        env.env_time == create_init_env_time() + passed_time + passed_time_2
    ), "Env time needs to be updated via the step function"


def test_time_limit():
    np.random.seed(42)
    env = Environment(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        ROOT_DIR / "game_content" / "item_info.yaml",
    )
    env.reset_env_time()

    assert not env.game_ended, "Game has not ended yet"

    passed_time = timedelta(seconds=10)
    env.step(passed_time)

    assert not env.game_ended, "Game has not ended yet"

    passed_time_2 = timedelta(
        seconds=(env.env_time_end - env.beginning_time).total_seconds()
    )
    env.step(passed_time_2)

    assert env.game_ended, "Game has ended now."
